#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef WIN32
#include <dlfcn.h>
#include <glob.h>
#endif

#include <libxml/parser.h>
#include <glib.h>

#include "binding.h"
#include "../common/xml_funcs.h"

#ifdef _PAPYRUS_GUI
#include "bc.h"
#else
#include "debug.h"
#endif

#ifdef WIN32
#include "binding_pg.h"
#endif

typedef struct
	{
	Binding *functions;
	char *libraryname;
	char *type;
	void *handle;
	}
Library;

GList *libraries = NULL;

#define ADD_BINDING(a)							\
	do {								\
		retval->a = dlsym (handle, #a );			\
		if (!retval->a) {					\
			errormsg ("Unable to load function " #a);	\
			free (retval);					\
			return NULL;					\
		}							\
	} while (0)

Binding *
build_bindings(void *handle)
	{
	Binding *retval;

	retval = malloc(sizeof * retval);
#ifndef WIN32
	ADD_BINDING(pap_doquery);
	ADD_BINDING(pap_query_reset);
	ADD_BINDING(pap_query_compare);
	ADD_BINDING(pap_query_ncolumns);
	ADD_BINDING(pap_query_column_name);
	ADD_BINDING(pap_query_length);
	ADD_BINDING(pap_query_next);
	ADD_BINDING(pap_query_has_more);
	ADD_BINDING(pap_query_get_value);
	ADD_BINDING(pap_connect_db);
	ADD_BINDING(pap_disconnect_db);
	ADD_BINDING(pap_clearquery);
	ADD_BINDING(pap_cleanup);
	ADD_BINDING(pap_init);
#endif
	retval->pap_init();

	return retval;
	}

/*
  Return: NULL if database node's type is not found.
*/
static char *
get_dbtype(xmlDocPtr doc)
	{
	xmlNodePtr node = NULL, curr = NULL;
	char *dbtype = NULL;

	curr = xmlDocGetRootElement(doc);
	if (curr == NULL)
	     {
	     fprintf(stderr,"Failed to read XML file.\n");
	     return NULL;
	     }
	   
	for (curr = curr->children; curr; curr = curr->next)
		{
		if (strcmp(curr->name, "database") == 0)
			{
			node = curr;
			}
		}

	if (node)
		dbtype = xmlGetProp(node, "type");

	/* printf ("Database Type: %s\n", (dbtype) ? dbtype : "Error"); */

	return dbtype;

	}

/*
  Francis: Currently, this stupid function exists
  only for the postgres -> pg mapping...
*/
static char *
get_postfix(char *dbtype)
	{
	char *retval = NULL;

	if (strcmp(dbtype, "postgres") == 0)
		retval = "pg";
	else if (strcmp(dbtype, "mysql") == 0)
		retval = "mysql";
	else if (strcmp(dbtype, "bonddb") == 0)
		retval = "bonddb";
	else if (strcmp(dbtype, "libgda") == 0)
		retval = "gda";

	return retval;
	}

static int
add_library(char *library)
	{
	Library *l;
	void *handle;
	char *(*func) (void);

	if (!library)
		return -1;
#ifndef WIN32
	handle = dlopen(library, RTLD_LAZY);
	if (!handle)
		{
		errormsg("%s", dlerror());
		return -1;
		}

	func = dlsym(handle, "pap_type");
	if (!func)
		{
		errormsg("%s, %s", dlerror(), library);
		errormsg("This is normally caused by not running ./configure "
		         "properly at compile time or this database backend "
		         "is not supported. Try running ./configure --enable-postgresql=yes");
		dlclose(handle);
		return -1;
		}

	l = malloc(sizeof * l);

	l->handle = handle;
	l->libraryname = strdup(library);
	l->type = strdup(func());
	l->functions = build_bindings(handle);
	if (!l->functions)
		{
		free(l->type);
		free(l->libraryname);
		free(l);
		dlclose(handle);
		return -1;
		}

	libraries = g_list_append(libraries, l);
#endif
	return 0;
	}

static int
add_libraries(GList * dirs, char *dbsuffix)
	{
	GList *cur;
	char *dir;
	glob_t g;
	char *match;
	int i;

	for (cur = dirs; cur != NULL; cur = cur->next)
		{
		dir = cur->data;
		asprintf(&match, "%s/libpapyrus_binding_%s.so", dir, dbsuffix);
		/* printf ("Pattern looking for: %s\n", match); */
		glob(match, 0, NULL, &g);

		for (i = 0; i < g.gl_pathc; i++)
			{
			add_library(g.gl_pathv[i]);
			}

		globfree(&g);
		}

	return 0;
	}

Binding *
binding_find(char *type)
	{
	GList *cur;
	Library *l;

	for (cur = libraries; cur != NULL; cur = cur->next)
		{
		l = cur->data;
		if (strcmp(l->type, type) == 0)
			return l->functions;
		}

	return NULL;
	}

int
binding_init(xmlDocPtr doc)
	{
	GList *bindingdirs = NULL;
	char *dbtype, *dbsuffix = NULL;

	if (libraries)
		{
		errormsg("Can't call twice");
		return -1;
		}

	/* Francis: only if getenv() returns non-NULL values */
	if (getenv("PAPYRUSLIBS"))
		{
		bindingdirs = g_list_append(bindingdirs, getenv("PAPYRUSLIBS"));
		}
	bindingdirs = g_list_append(bindingdirs, ".");
	bindingdirs = g_list_append(bindingdirs, "./.libs");
	bindingdirs = g_list_append(bindingdirs, "./backend");
	bindingdirs = g_list_append(bindingdirs, "./backend/.libs");
	bindingdirs = g_list_append(bindingdirs, "/usr/local/lib");
	bindingdirs = g_list_append(bindingdirs, "/usr/lib");

	/* Get database manager type: e.g. postgres or mysql */
	dbtype = get_dbtype(doc);
	/* Default binding to postgres (= pg) */
	dbsuffix = (dbtype) ? get_postfix(dbtype) : "pg";

	add_libraries(bindingdirs, dbsuffix);

	return 0;
	}

int
binding_cleanup(void)
	{
	GList *cur;
	Library *l;

	for (cur = libraries; cur != NULL; cur = cur->next)
		{
		l = cur->data;
		l->functions->pap_cleanup();
		free(l->libraryname);
		free(l->type);
#ifndef WIN32
		dlclose(l->handle);
#endif
		free(l->functions);
		free(l);
		}

	g_list_free(libraries);
	libraries = NULL;

	return 0;
	}
