#define _GNU_SOURCE
#ifdef USE_LIBGDA

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include <libgda/libgda.h>

#define errormsg(fmt,args...) fprintf (stderr, "%s:%d:%s(): " fmt "\n", __FILE__, __LINE__, __PRETTY_FUNCTION__ , ##args)

#include "binding_generic.h"

typedef struct {
	GdaDataModel *res;
	GdaConnection *conn;
	gint cur_record;
	gchar *string;
} gda_query;

void *
pap_doquery(void *connection, gchar *query)
{
	gda_query *result;

	GdaCommand *command;
	GdaConnection *conn = connection;
	GdaDataModel *dm;

	if (!conn) {
		errormsg("Invalid connection");
		return NULL;
	}

	if (!query || *query == '\0') {
		errormsg("Invalid query string");
		return NULL;
	}

	command = gda_command_new(query, 
				  GDA_COMMAND_TYPE_SQL, 
				  GDA_COMMAND_OPTION_STOP_ON_ERRORS);

	dm = gda_connection_execute_single_command(conn, command, NULL);

	if (!dm) {
		g_error("LibGda error returning data model from SQL query");
		return NULL;
	}

	result = malloc(sizeof(*result));

	result->conn = conn;
	result->cur_record = 0;
	result->res = dm;
	result->string = strdup(query);

	return result;
}

gint
pap_query_reset(void *query)
{
	gda_query *q = query;

	q->cur_record = 0;

	return 0;
}

gint
pap_query_compare(void *query, gchar *querystr)
{
	gda_query *q = query;

	return strcmp(q->string, querystr);
}

gint
pap_query_ncolumns(void *query)
{
	gda_query *q = query;
	return gda_data_model_get_n_columns(q->res);
}

gchar *
pap_query_column_name(void *query, gint pos)
{
	gda_query *q = query;
	return (gchar *)gda_data_model_get_column_title(q->res, pos);
}

gint
pap_query_length(void *query)
{
	gda_query *q = query;
	return gda_data_model_get_n_rows(q->res);
}

gint
pap_query_next(void *query)
{
	gda_query *q = query;

	/*	errormsg ("string: %s, cur_record: %d", q->string, q->cur_record); */

	if (pap_query_length(query) <= 0)
		return -1;

	if (q->cur_record == pap_query_length(query) - 1)
		return -1;
 
	q->cur_record++;

	return 0;
}

gint
pap_query_has_more(void *query)
{
	gda_query *q = query;

	if (!query)
		return 0;

	if (pap_query_length(query) <= 0)
		return 0;

	return (q->cur_record <= (pap_query_length(query) - 1));
}

gchar *
pap_query_get_value(void *query, gchar *field)
{
	gda_query *q = query;
	gint fieldno;
	gchar *retval;
	GdaValue *value;

	if (q == NULL || q->res == NULL) {
		errormsg("Invalid query recordset");
		return NULL;
	}

	if (pap_query_length(q) <= 0) {
		errormsg("No rows returned by query %s", q->string);
		return NULL;
	}

	fieldno = gda_data_model_get_column_position(q->res, field);
	if (fieldno == -1) {
		errormsg("Invalid field: %s in query %s", field, q->string);
		return NULL;
	}

	value = (GdaValue *)gda_data_model_get_value_at(q->res, fieldno, q->cur_record);
	if (!value) {
		g_error("Failed to get value at row %d, col %d", q->cur_record, fieldno);
		return NULL;
	}
	retval = gda_value_stringify(value);
	if (!retval) {
		errormsg("Error getting value `%s' in query `%s'", field, q->string);
		return NULL;
	}

	g_strstrip(retval);
	if (!strcmp(retval, "NULL"))
	    retval = "";
	return retval;
}

void *
pap_connect_db(char *connectstr, char *provider)
{
	GdaConnection *conn;
	GdaClient *client;
	gchar *newconnectstr;

	client = gda_client_new();
	
        newconnectstr = g_strdelimit(connectstr, " \t\r\n", ' ');
	newconnectstr = g_strstrip(newconnectstr);

	g_message("Connecting to database...");
	conn = gda_client_open_connection_from_string(client,
						      provider,
						      newconnectstr,
						      GDA_CONNECTION_OPTIONS_READ_ONLY);

	if (!GDA_IS_CONNECTION(conn)) {
		g_error("Could not open connection from string `%s', using provider `%s'", 
			connectstr, provider);
		return NULL;
	}
	else
		g_message("Connected");
	
	return conn;
}

gint
pap_disconnect_db(void *connection)
{
	GdaConnection *conn = connection;
	GdaClient *client;

	client = gda_connection_get_client(conn);
	gda_client_close_all_connections(client);

	return 0;
}

gint
pap_clearquery(void *query)
{
	gda_query *q = query;

	if (!q || !q->res) {
		errormsg("Invalid query");
		return -1;
	}

	g_free(q->string);

	/* not sure how to free GdaDataModel */

	g_free(q);

	return 0;
}

gchar *
pap_type()
{
	static gchar *retval = NULL;

	if (!retval)
		retval = strdup("libgda");

	return retval;
}


/* Wouldn't it be nice if these were implemented? */
gint
pap_init()
{
	gda_init ("Papyrus", "0.9", 0, NULL);
	return 0;
     
}

gint
pap_cleanup()
{
	return 0;
}

#endif
