#define _GNU_SOURCE
#ifdef USE_MYSQL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include <mysql/mysql.h>

#define errormsg(fmt,args...) fprintf (stderr, "%s:%d:%s(): " fmt "\n", __FILE__, __LINE__, __PRETTY_FUNCTION__ , ##args)

#include "binding_generic.h"

typedef struct
{
        MYSQL *conn;
        MYSQL_RES *res;
        int cur_record;
        char *string;
}
mysql_query_type;

void *
pap_doquery(void *connection, char *query)
{
        mysql_query_type *retval;
        MYSQL *conn = connection;
        MYSQL_RES *res;

        if (!conn) {
                errormsg("Invalid connection");
                return NULL;
        }

        if (!query || *query == '\0') {
                errormsg("Invalid query string");
                return NULL;
        }

        /* According to MYSQL C API documentation, the query string should not end
           with ';' or '\g' */

        if (mysql_real_query(conn, query, strlen(query))) {
                errormsg("Mysql error: %s", mysql_error(conn));
                return NULL;
        } else {
                res = mysql_store_result(conn);
        }

        retval = malloc(sizeof(*retval));

        retval->conn = conn;
        retval->cur_record = 0;
        retval->res = res;
        retval->string = strdup(query);

        return retval;
}

int
pap_query_reset(void *query)
{
        mysql_query_type *q = query;

        q->cur_record = 0;

        return 0;
}

int
pap_query_compare(void *query, char *querystr)
{
        mysql_query_type *q = query;

        return strcmp(q->string, querystr);
}

int
pap_query_ncolumns(void *query)
{
        mysql_query_type *q = query;

        return mysql_num_fields(q->res);
}

char *
pap_query_column_name(void *query, int pos)
{
        MYSQL_FIELD *field;
        mysql_query_type *q = query;

        field = mysql_fetch_field_direct(q->res, pos);
        g_assert(field);
        return field->name;
}

int
pap_query_length(void *query)
{
        mysql_query_type *q = query;

        return mysql_num_rows(q->res);
}

int
pap_query_next(void *query)
{
        mysql_query_type *q = query;

        /* errormsg ("string: %s, cur_record: %d", q->string, q->cur_record); */

        if (pap_query_length(query) <= 0)
                return -1;

        if (q->cur_record == pap_query_length(query) - 1)
                return -1;

        q->cur_record++;

        return 0;
}

int
pap_query_has_more(void *query)
{
        mysql_query_type *q = query;

        if (!query)
                return 0;

        if (pap_query_length(query) <= 0)
                return 0;

        return (q->cur_record <= (pap_query_length(query) - 1));
}

char *
pap_query_get_value(void *query, char *field)
{
        int i;
        int fieldno;
        char *retval;

        MYSQL_ROW row;
        MYSQL_FIELD *fields;
        int numfields;

        mysql_query_type *q = query;

        if (q == NULL || q->res == NULL) {
                errormsg("Invalid query recordset");
                return NULL;
        }

        if (pap_query_length(q) <= 0) {
                errormsg("No rows returned by query %s", q->string);
                return NULL;
        }

        /* debugmsg ("query: %s, field: %s", q->string, field); */

        numfields = mysql_num_fields(q->res);
        fields = mysql_fetch_fields(q->res);

        fieldno = -1;
        /* Francis: unfortunately, mysql C API doesn't have Postgresql's
           PQfnumber(), which takes field name in string and returns its number.
           WARNING: the folloing for-loop just return the very first match */
        for (i = 0; i < numfields; i++) {
                if (!strcmp(fields[i].name, field)) {
                        /* Set field number */
                        fieldno = i;
                        break;
                }
        }

        if (fieldno == -1) {
                errormsg("Invalid field: %s in query %s", field, q->string);
                return NULL;
        }

        mysql_data_seek(q->res, (unsigned long long)q->cur_record);
        row = mysql_fetch_row(q->res);
        retval = row[fieldno];

        /* Francis: Hmm, null value is also a valid value after all...
        	if (!retval)
        		errormsg ("Error getting value %s in query %s", field, q->string);
        */

        /* debugmsg ("[%d] query=%s, field=%s (%d) = %s", q->cur_record, q->string,
           field, fieldno, retval); */

        return retval;
}

void *
pap_connect_db(char *connectstr, char *provider)
{
        /* Francis: Mysql C API have different style from postgresql... */
        MYSQL *conn;
        gchar **conn_fields;
        gchar *newconnectstr;

        char *host = NULL;
        char *user = NULL;
        char *passwd = NULL;
        char *db = NULL;

        int i;

        /* This will be freed by when mysql_close() is called later WARING: If you
           want to allocate some memory space, you must pass NULL valued argument
           SEE: mysql_init() */
        conn = mysql_init(NULL);

        newconnectstr = g_strstrip(connectstr);
        /* Francis: I am worried about the delimeters used in the connection string.
           Hope this eliminate any problems */
        newconnectstr = g_strdelimit(connectstr, " \t\r\n", ' ');

        /* Francis: I think 4 is enough (db, host, user, passwd) */
#define MAX_TOKEN 4

        conn_fields = g_strsplit(newconnectstr, " ", MAX_TOKEN);

        for (i = 0; i < MAX_TOKEN && conn_fields[i]; i++) {
                /* printf("conn_fields[%d]:%s\n", i, conn_fields[i]); */
                g_strchomp(conn_fields[i]);

                /* if host is null, then assign a string position if "host=" is found */
                if (!host && (host = strstr(conn_fields[i], "host="))) {
                        host += strlen("host=");
                } else if (!db && (db = strstr(conn_fields[i], "dbname="))) {
                        db += strlen("dbname=");
                } else if (!user && (user = strstr(conn_fields[i], "user="))) {
                        user += strlen("user=");
                } else if (!passwd && (passwd = strstr(conn_fields[i], "passwd="))) {
                        passwd += strlen("passwd=");
                }
        }

        /* printf("host:%s\nuser:%s\npasswd:%s\ndb:%s\n", host,user,passwd,db); */
        /* port; use default port number by inspecting host */
        /* unix_socket; use appropriate socket by looking at host */
        /* client_flag */
        if (!mysql_real_connect(conn, host, user, passwd, db, 0, NULL, 0)) {
                errormsg("Mysql: failed to connect to database: Error :%s",
                         mysql_error(conn));
                return NULL;
        }

        return conn;
}

int
pap_disconnect_db(void *connection)
{
        MYSQL *conn = connection;

        mysql_close(conn);

        return 0;
}

int
pap_clearquery(void *query)
{
        mysql_query_type *q = query;

        if (!q || !q->res) {
                errormsg("Invalid query");
                return -1;
        }

        free(q->string);
        mysql_free_result(q->res);

        free(q);

        return 0;
}

char *
pap_type()
{
        static char *retval = NULL;

        if (!retval)
                retval = strdup("mysql");

        return retval;
}

int
pap_init()
{
        return 0;
}

int
pap_cleanup()
{
        return 0;
}

#endif
