#define _GNU_SOURCE
#ifdef USE_POSTGRESQL

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include <postgres_ext.h>
#include <libpq-fe.h>

#define errormsg(fmt,args...) fprintf (stderr, "%s:%d:%s(): " fmt "\n", __FILE__, __LINE__, __PRETTY_FUNCTION__ , ##args)

#include "binding_generic.h"

typedef struct
{
        PGresult *res;
        PGconn *conn;
        int cur_record;
        char *string;
}
	postgres_query;

void *
pap_doquery(void *connection, char *query)
{
        postgres_query *retval;
        PGconn *conn = connection;
        PGresult *res;

        /* Francis: To get better debuging message, added status checking */
        ExecStatusType status;

        if (!conn) {
                errormsg("Invalid connection");
                return NULL;
        }

        if (!query || *query == '\0') {
                errormsg("Invalid query string");
                return NULL;
        }

        res = PQexec(conn, query);

        if (!res) {
                errormsg("Postgres error: %s", PQerrorMessage(conn));
                return NULL;
        }

        /* Francis: Added following checking */
        status = PQresultStatus(res);

        switch (status) {
        case PGRES_EMPTY_QUERY:
        case PGRES_COMMAND_OK:
        case PGRES_TUPLES_OK:
                break;

        case PGRES_COPY_OUT:
        case PGRES_COPY_IN:
        case PGRES_BAD_RESPONSE:
        case PGRES_NONFATAL_ERROR:
        case PGRES_FATAL_ERROR:
                errormsg("Postgres error: %s", PQresultErrorMessage(res));
                return NULL;
        }
        /* Francis: Modification ended here */

        retval = malloc(sizeof(*retval));

        retval->conn = conn;
        retval->cur_record = 0;
        retval->res = res;
        retval->string = strdup(query);

        return retval;
}

int
pap_query_reset(void *query)
{
        postgres_query *q = query;

        q->cur_record = 0;

        return 0;
}

int
pap_query_compare(void *query, char *querystr)
{
        postgres_query *q = query;

        return strcmp(q->string, querystr);
}

int
pap_query_ncolumns(void *query)
{
        postgres_query *q = query;

        return PQnfields(q->res);
}

char *
pap_query_column_name(void *query, int pos)
{
        postgres_query *q = query;

        return PQfname(q->res, pos);
}

int
pap_query_length(void *query)
{
        postgres_query *q = query;

        return PQntuples(q->res);
}

int
pap_query_next(void *query)
{
        postgres_query *q = query;

        /*	errormsg ("string: %s, cur_record: %d", q->string, q->cur_record); */

        if (pap_query_length(query) <= 0)
                return -1;

        if (q->cur_record == pap_query_length(query) - 1)
                return -1;

        q->cur_record++;

        return 0;
}

int
pap_query_has_more(void *query)
{
        postgres_query *q = query;

        if (!query)
                return 0;

        if (pap_query_length(query) <= 0)
                return 0;

        return (q->cur_record <= (pap_query_length(query) - 1));
}

char *
pap_query_get_value(void *query, char *field)
{
        postgres_query *q = query;
        int fieldno=-1;
        char *retval = NULL;

        if (q == NULL || q->res == NULL) {
                errormsg("Invalid query recordset");
                return NULL;
        }

        if (pap_query_length(q) <= 0) {
			/*errormsg("No rows returned by query %s", q->string); */
                return NULL;
        }

        /* debugmsg ("query: %s, field: %s", q->string, field); */

        fieldno = PQfnumber(q->res, field);

        if (fieldno == -1) {
                errormsg("Invalid field: %s in query %s", field, q->string);
                return NULL;
        }

        retval = PQgetvalue(q->res, q->cur_record, fieldno);

        if (!retval)
                errormsg("Error getting value %s in query %s", field, q->string);

        /* debugmsg ("[%d] query=%s, field=%s (%d) = %s", q->cur_record, q->string, field, fieldno, retval); */
        g_strstrip(retval);
        return retval;
}

void *
pap_connect_db(char *connectstr, char *provider)
{
        PGconn *conn;

        conn = PQconnectdb(connectstr);

        if (!conn)
                return NULL;

        if (PQstatus(conn) == CONNECTION_BAD) {
                PQfinish(conn);
                return NULL;
        }

        return conn;
}

int
pap_disconnect_db(void *connection)
{
        PGconn *conn = connection;

        PQfinish(conn);

        return 0;
}

int
pap_clearquery(void *query)
{
        postgres_query *q = query;

        if (!q || !q->res) {
                errormsg("Invalid query");
                return -1;
        }

        free(q->string);
        PQclear(q->res);
        free(q);
        return 0;
}

char *
pap_type()
{
        static char *retval = NULL;
        if (!retval)
                retval = strdup("postgres");

        return retval;
}

int
pap_init()
{
        return 0;
}

int
pap_cleanup()
{
        return 0;
}

#endif
