#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include <glib.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#include "debug.h"
#include "xml_funcs.h"
#include "datasource.h"
#include "code.h"
#include "binding.h"
#include "convert.h"

#ifdef WIN32
#include "binding_pg.h"
#endif

#include "equation.h"

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

static int convert_table(xmlNodePtr report, FILE * output, xmlNodePtr table);
static int convert_title(xmlNodePtr report, FILE * output, xmlNodePtr title);
int convert_datasource(xmlNodePtr report, FILE * output, xmlNodePtr datasource);

static int convert_dataloop(xmlNodePtr report, FILE * output, xmlNodePtr loop);
static int convert_header(xmlNodePtr report, FILE * output, xmlNodePtr header);
static int convert_clearpage(xmlNodePtr report, FILE * output, xmlNodePtr node);
static int convert_newpage(xmlNodePtr report, FILE * output, xmlNodePtr node);

static int
print_tag(FILE * output, const char *tagname, xmlNodePtr node, char **ignore,
          int len)
	{
	xmlAttr *attr;
	char *text, *tmpstr;
	int i, count = 0;
	if (!node)
	     return -1;
	
	fprintf(output, "<%s", tagname);
	for (attr = node->properties; attr != NULL; attr = attr->next)
		{
		text = attr->children->content;
		for (i = 0; i < len; i++)
			{
			if (strcmp(ignore[i], attr->name) == 0)
				text = NULL;
			}

		if (text)
			{
			tmpstr = g_markup_escape_text(text,strlen(text));
			fprintf(output, " %s=\"%s\"", attr->name, tmpstr);
			g_free(tmpstr);
			count++;
			}
		}

	fprintf(output, ">");

	return count;
	}

static int
convert_string(FILE * output, char *string)
	{
	char *tmpstr;

	if (!string)
		return -1;

#ifdef OBSOLETE
	/* Tim: Removing isolated comma from the start of the string */
	if (*string == ' ')
		string++;
	for (i=strlen(string)-1;i>=0;i++)
	     {	
	     if (string[i] == ' ' || string[i] == 0 || 
		 string[i] == '\n' || string[i] == '\t')
	       string[i] = 0;
	     else
		  break;
	     }
	   

	   /* End */
	for (c = string; *c != '\0'; c++)
		{
		switch (*c)
			{
			/* Tim: Removing comma at end of string */
		case ',':
			if (*(c + 1) != '\0')
				fprintf(output, "%c", *c);
			break;
			/* End */
		default:
			fprintf(output, "%c", *c);
			}
		}
#endif
	   
	tmpstr = g_markup_escape_text(string,strlen(string));
	fprintf(output,"%s",tmpstr);
	g_free(tmpstr);
	return 0;
	}

/*
  Francis: convert node 'arg'
*/
static int
convert_special_arg(FILE * output, xmlNodePtr node)
	{

	xmlNodePtr curr;

	print_tag(output, node->name, node, NULL, 0);

	for (curr = node->children; curr; curr = curr->next)
		{
		if (strcmp(curr->name, "text") == 0)
			{
			convert_string(output, curr->content);
			}
		else
			{
			errormsg("Invalid arg item: %s", curr->name);
			}
		}

	fputs("</arg>\n", output);

	return 0;
	}

/*
  Francis: I think this will be handy
*/
static int
convert_special(FILE * output, xmlNodePtr node)
	{
	xmlNodePtr curr;

	print_tag(output, node->name, node, NULL, 0);
	fputs("\n", output);

	for (curr = node->children; curr; curr = curr->next)
		{
		if (strcmp(curr->name, "arg") == 0)
			{
			convert_special_arg(output, curr);
			}
		else if (strcmp(curr->name, "text") == 0)
			;
		else if (strcmp(curr->name, "comment") == 0)
			;
		else
			{
			errormsg("Invalid special item: %s", curr->name);
			}
		}

	fputs("</special>\n", output);

	return 0;
	}

static DataSource *
node_get_datasource(xmlNodePtr node)
	{
	DataSource *retval;
	char *datasource_name;

	/* Francis: extended for more complex expression */
	char *dataloop_name;

	datasource_name = xmlGetProp(node, "source");
	dataloop_name = xmlGetProp(node, "loopref");

	if (datasource_name)
		{
		retval = datasource_get(datasource_name);
		if (!retval)
			errormsg("Invalid datasource: %s", datasource_name);
		}
	else
		{
		datasource_name = xml_node_parent_attr(node, "source");
		if (datasource_name)
			{
			retval = datasource_get(datasource_name);
			if (!retval)
				errormsg("Invalid datasource: %s", datasource_name);
			}
		else
			retval = NULL;
		}

	return retval;
	}

static char *
data_retrieve(xmlNodePtr data)
	{
	char *variable = NULL, *retval;
	DataSource *datasource;
	xmlNodePtr item;
	char *type;
	char buffer[64];
	unsigned long long days;
	GDate *date;

	char *nulltozero;
	char *nulltospace;
	char *loopref;

	/* Francis: Check if nulltozero attribute exist */
	nulltozero = xmlGetProp(data, "nulltozero");
	nulltospace = xmlGetProp(data, "nulltospace");

	/* Francis: Look for dataloop */
	loopref = xmlGetProp(data, "loopref");

	datasource = node_get_datasource(data);

	for (item = data->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "text") == 0)
			{
			if (variable)
				errormsg("Cannot have more than one entry in a data item");
			else
				variable = item->content;
			}
		else
			errormsg("Invalid data item: %s", item->name);
		}

	retval = variable_process_full(datasource, variable);

	/* Francis: Hello there ! */
	if (nulltozero != NULL)
		{
		if (strcmp(nulltozero, "true") == 0)
			{
			if (retval == NULL || *retval == '\0')
				{
				retval = strdup("0");
				}
			}
		}

	if (nulltospace != NULL)
		{
		if (strcmp(nulltospace, "true") == 0)
			{
			if (retval == NULL || *retval == '\0')
				{
				retval = strdup(" ");
				}
			}
		}

	type = xmlGetProp(data, "type");

	if (type && strcmp(type, "date") == 0)
		{
		if (retval && (*retval != '\0') && (strcmp(retval, "0001-01-01") != 0))
			{
			date = g_date_new();
			g_date_set_parse(date, retval);	/* heurisitic to guess date and
																			   put into date variable */

			if (g_date_valid(date))
				{

				days = g_date_get_julian(date);	/* get the 'julian' date
																						   from date variable */

				g_date_free(date);	 /* free up date variable */

				sprintf(buffer, "%llu", days);

				/* errormsg ("buffer is: %s from %s", buffer, retval); */

				free(retval);
				retval = strdup(buffer);
				}
			else
				errormsg("Date not valid");
			}
		else if (retval)
			{
			free(retval);
			retval = strdup("NaN");
			}
		}
	return retval;
	}

static int
convert_data(xmlNodePtr report, FILE * output, xmlNodePtr data)
	{
	char *retval;

	retval = data_retrieve(data);

	convert_string(output, retval);

	free(retval);

	return 0;
	}

static char *convert_equation_str(xmlNodePtr equation);

static long double
convert_equation_single(xmlNodePtr equation)
	{
	xmlNodePtr child;
	char *old, *eq_str = NULL;
	char *data;
	long double retval = INFINITY;

	for (child = equation->children; child != NULL; child = child->next)
		{
		old = eq_str;
		if (strcmp(child->name, "text") == 0)
			{
			eq_str = string_extend(eq_str, child->content);
			free(old);
			}
		else if (strcmp(child->name, "data") == 0)
			{
			data = data_retrieve(child);
			eq_str = string_extend(eq_str, data);
			free(data);
			free(old);
			}
		else if (strcmp(child->name, "equation") == 0)
			{
			data = convert_equation_str(child);
			eq_str = string_extend(eq_str, data);
			free(data);
			free(old);
			}
		else
			errormsg("Invalid equation item: %s", child->name);
		}
	retval = equation_evaluate(eq_str);
        /* errormsg("%s return: %s",eq_str, retval); */

	free(eq_str);

	return retval;
	}

typedef enum
{
    EQN_none,
    EQN_sum,
    EQN_min,
    EQN_max,
    EQN_average
} equation_type;

static char *
convert_equation_str(xmlNodePtr equation)
	{
	long double eq, total = 0;
	char *format, *full_format, *type, *datasource, *retval, *truncate;
	DataSource *data;
	equation_type eqntype;
	int count = 0;
	int trunc;

	type = xmlGetProp(equation, "type");
	datasource = xmlGetProp(equation, "source");
	format = xmlGetProp(equation, "format");

	if (type)
		{
		if (strcmp(type, "sum") == 0)
			eqntype = EQN_sum;
		else if (strcmp(type, "maximum") == 0)
			eqntype = EQN_max;
		else if (strcmp(type, "minimum") == 0)
			eqntype = EQN_min;
		else if (strcmp(type, "average") == 0)
			eqntype = EQN_average;
		else
			{
			errormsg("Invalid equation type: %s", type);
			eqntype = EQN_none;

			}
		}
	else
		eqntype = EQN_none;

	if (eqntype != EQN_none)
		{
		if (!datasource)
			errormsg("Must specify 'source' with 'type' on equation");

		data = datasource_get(datasource);

		if (!data)
			{
			errormsg("Invalid equation: Unable to find datasource: %s",
			         datasource);
			return "";
			}

		/* reset the row so its on position 0 */

		datasource_query(data);
		}
	else
		{
		data = NULL;
		datasource = NULL;
		}

	if (datasource && data)
		{
#ifndef WIN32
		if (data->db->binding->pap_query_length(data->db_query) > 0)
#else
		if (pap_query_length(data->db_query) > 0)
#endif		     
		     {
			count = 0;
			if (eqntype == EQN_min)
				total = INFINITY;
			while (1)
				{
				eq = convert_equation_single(equation);
				if (finite(eq))		 /* Was: !isnan (eq) && !isinf (eq) */
					{
					/* errormsg ("Added %Lf", eq); */
					switch (eqntype)
						{
					case EQN_sum:
						total += eq;
						break;
					case EQN_max:
						total = max(total, eq);
						break;
					case EQN_min:
						total = min(total, eq);
						break;
					case EQN_average:
						total += eq;
						count++;

						break;
					case EQN_none:
						total += eq;
						break;
						}
					}
#ifndef WIN32
				if (data->db->binding->pap_query_next(data->db_query) != 0)
					break;
#else
				if (pap_query_next(data->db_query) != 0)
					break;
#endif				
				}
			if (eqntype == EQN_average)
				{
				total = total / count;
				}

			}
		}
	else
		total = convert_equation_single(equation);

	truncate = xmlGetProp(equation, "truncate");

	if (truncate && ((total + 0.000001 <= 0) || (total - 0.000001 > 0)))
		{
		double val;

		trunc = atoi(truncate);
		val = 0.5 * (pow(10, -trunc));
		total -= val;
		}

	full_format = g_strdup_printf("%%%sLf", format ? format : "");
	retval = g_strdup_printf(full_format,total);
	g_free(full_format);

	return retval;
	}

static int
convert_equation(xmlNodePtr report, FILE * output, xmlNodePtr equation)
	{
	char *eqn, *tmpstr;

	eqn = convert_equation_str(equation);
	
	tmpstr = g_markup_escape_text(eqn,strlen(eqn));
	fprintf(output, "%s", tmpstr);
	g_free(tmpstr);
	free(eqn);

	return 0;
	}

static int
convert_generic(xmlNodePtr report, FILE * output, xmlNodePtr node)
	{
	xmlNodePtr item;

	/* char *trimtext; */

	print_tag(output, node->name, node, NULL, 0);

	for (item = node->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "text") == 0)
			{
			/* trimtext = trim (item->content); */
			convert_string(output, item->content);
			/* fprintf (output, "%s", trimtext); free (trimtext); */
			}
		else if (strcmp(item->name, "data") == 0)
			convert_data(report, output, item);
		else if (strcmp(item->name, "equation") == 0)
			convert_equation(report, output, item);
		else
			convert_generic(report, output, item);
		}

	fprintf(output, "</%s>", node->name);

	return 0;
	}

static int
convert_cell(xmlNodePtr report, FILE * output, xmlNodePtr cell)
	{
	xmlNodePtr item;

	print_tag(output, "cell", cell, NULL, 0);

	for (item = cell->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0 ||
		        strcmp(item->name, "paragraph") == 0 ||
		        strcmp(item->name, "image") == 0)
			convert_generic(report, output, item);
		else if (strcmp(item->name, "table") == 0)
			convert_table(report, output, item);
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid cell item: %s", item->name);
		}

	fprintf(output, "</cell>\n");

	return 0;
	}

static int
convert_row(xmlNodePtr report, FILE * output, xmlNodePtr row)
	{
	xmlNodePtr loopitem;

	/* ignore is a list of keywords to ignore when outputing the xml */

	print_tag(output, "row", row, NULL, 0);

	fprintf(output, "\n");
	/* print out every loopitem */
	for (loopitem = row->children; loopitem != NULL; loopitem = loopitem->next)
		{
		if (strcmp(loopitem->name, "cell") == 0)
			convert_cell(report, output, loopitem);
		else if (strcmp(loopitem->name, "dataloop") == 0)
			convert_dataloop(report, output, loopitem);
		else if (strcmp(loopitem->name, "table") == 0)
			convert_table(report, output, loopitem);
		else if (strcmp(loopitem->name, "datasource") == 0)
			convert_datasource(report, output, loopitem);
		else if (strcmp(loopitem->name, "text") == 0)
			;
		else if (strcmp(loopitem->name, "comment") == 0)
			;
		else 	
		     convert_generic(report->children, output, loopitem);
		}
	fprintf(output, "</row>\n");

	return 0;
	}

static int
convert_dataloop(xmlNodePtr report, FILE * output, xmlNodePtr loop)
	{
	xmlNodePtr item;
	char *datasource = NULL;
	DataSource *data;

	datasource = xmlGetProp(loop, "source");

	if (datasource)
		{
		data = datasource_get(datasource);

		if (!data)
			{
			errormsg("Invalid dataloop: Unable to find datasource: %s",
			         datasource);
			return -1;
			}

		/* reset the row so it's on position 0 */
		datasource_query(data);
		}

	if (!datasource || !data || !data->db || !data->db_query)
		{
		errormsg("Invalid dataloop: no source");
		return -1;
		}

#ifndef WIN32
	   if (data->db->binding->pap_query_length(data->db_query) > 0)
#else
	   if (pap_query_length(data->db_query) > 0)
#endif	     
	        {
		while (1)
			{
			/* ignore is a list of keywords to ignore when outputing the xml
			 */

			for (item = loop->children; item != NULL; item = item->next)
				{
				if (strcmp(item->name, "row") == 0)
					convert_row(report, output, item);
				else if (strcmp(item->name, "dataloop") == 0)
					convert_dataloop(report, output, item);
				else if (strcmp(item->name, "datasource") == 0)
					convert_datasource(report, output, item);
				else if (strcmp(item->name, "text") == 0)
					;
				else if (strcmp(item->name, "comment") == 0)
					;
				else if (strcmp(item->name, "cell") == 0)
					convert_cell(report, output, item);
				else if (strcmp(item->name, "table") == 0)
					convert_table(report, output, item);
				else if (strcmp(item->name, "header") == 0)
					convert_header(report, output, item);
				else if (strcmp(item->name, "clearpage") == 0)
					convert_clearpage(report, output, item);
				else if (strcmp(item->name, "newpage") == 0)
					convert_newpage(report, output, item);
				else
					errormsg("Invalid dataloop item: %s", item->name);
				}
#ifndef WIN32
			if (data->db->binding->pap_query_next(data->db_query) != 0)
				break;
#else
			if (pap_query_next(data->db_query) != 0)
				break;
#endif			
			}
		}

	return 0;
	}

static int
convert_title(xmlNodePtr report, FILE * output, xmlNodePtr title)
	{
	xmlNodePtr item;

	print_tag(output, "title", title, NULL, 0);

	/* Make sure title tag has correct syntax */
	for (item = title->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0)
			convert_generic(report, output, item);
		else
			errormsg("Invalid title item: %s", item->name);
		}

	fprintf(output, "</title>\n");
	return 0;
	}

static int
convert_header(xmlNodePtr report, FILE * output, xmlNodePtr header)
	{
	xmlNodePtr item;

	print_tag(output, "header", header, NULL, 0);

	for (item = header->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			convert_row(report, output, item);
		else if (strcmp(item->name, "datasource") == 0)
			convert_datasource(report, output, item);
		else if (strcmp(item->name, "dataloop") == 0)
			convert_dataloop(report, output, item);
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid header item: %s", item->name);
		}

	fprintf(output, "</header>\n");

	return 0;
	}

static int
convert_footer(xmlNodePtr report, FILE * output, xmlNodePtr footer)
	{
	xmlNodePtr item;

	print_tag(output, "footer", footer, NULL, 0);

	for (item = footer->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			convert_row(report, output, item);
		else if (strcmp(item->name, "dataloop") == 0)
			convert_dataloop(report, output, item);
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid footer item: %s", item->name);
		}

	fprintf(output, "</footer>\n");

	return 0;
	}

static int
convert_autotable(xmlNodePtr report, FILE * output, xmlNodePtr autotable)
	{
	int ncolumns, i;
	char **columns;
	char *datasource, *str;
	char *border, *tmpstr;
	DataSource *data;

#ifdef PAPYRUS_EVALUATION

	int nrecords = 0;
#endif

	datasource = xmlGetProp(autotable, "source");
	border = xmlGetProp(autotable, "border");

	if (datasource)
		{
		data = datasource_get(datasource);

		if (!data)
			{
			errormsg("Invalid autotable: Unable to find datasource: %s",
			         datasource);
			return -1;
			}

		/* reset the row so its on position 0 */
		datasource_query(data);
		}

	if (!datasource || !data || !data->db || !data->db_query)
		{
		errormsg("Invalid autotable: no datasource");
		return -1;
		}
#ifndef WIN32
	ncolumns = data->db->binding->pap_query_ncolumns(data->db_query);
#else
	ncolumns = pap_query_ncolumns(data->db_query);
#endif	   
	if (ncolumns <= 0)
		{
		errormsg("Query %s has no columns", data->query);
		return -1;
		}

	columns = malloc(ncolumns * sizeof(char *));

#ifndef WIN32
	   for (i = 0; i < ncolumns; i++)
		columns[i] = data->db->binding->pap_query_column_name(data->db_query, i);
#else
	   for (i = 0; i < ncolumns; i++)
		columns[i] = pap_query_column_name(data->db_query, i);
#endif	   
	fprintf(output, "<table");
	if (border)
		fprintf(output, " border=\"%s\"", border);

	fprintf(output, ">\n");

	fprintf(output, "<header>\n<row>\n");
	for (i = 0; i < ncolumns; i++)
		{
		fprintf(output, "<cell><label bold=\"true\">%s</label></cell>\n",
		        columns[i]);
		}
	fprintf(output, "</row>\n</header>\n");
#ifndef WIN32
	if (data->db->binding->pap_query_length(data->db_query) > 0)
#else
	if (pap_query_length(data->db_query) > 0)
#endif
	     {
		while (1)
			{			
			fprintf(output, "<row>\n");
			for (i = 0; i < ncolumns; i++)
				{
				str = variable_process_full(data, columns[i]);
				tmpstr = g_markup_escape_text(str,strlen(str));
				   /* Tim: Removing isolated comma */
				if (strcmp(",", str) != 0)
					fprintf(output, "<cell><label>%s</label></cell>", tmpstr);
				g_free(tmpstr);
				free(str);
				}
			fprintf(output, "</row>\n");
#ifndef WIN32			   
			if (data->db->binding->pap_query_next(data->db_query) != 0)
				break;
#else
			if (pap_query_next(data->db_query) != 0)
				break;
#endif			
			}
		}

	fprintf(output, "</table>\n");

	free(columns);

	return 0;
	}

static int
convert_table(xmlNodePtr report, FILE * output, xmlNodePtr table)
	{
	xmlNodePtr item;

	print_tag(output, "table", table, NULL, 0);

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			convert_row(report, output, item);
		else if (strcmp(item->name, "datasource") == 0)
			convert_datasource(report, output, item);
		else if (strcmp(item->name, "header") == 0)
			convert_header(report, output, item);
		else if (strcmp(item->name, "footer") == 0)
			convert_footer(report, output, item);
		else if (strcmp(item->name, "dataloop") == 0)
			convert_dataloop(report, output, item);
		else if (strcmp(item->name, "line") == 0)
			convert_generic(report, output, item);
		else if (strcmp(item->name, "text") == 0)
			;
		else if (strcmp(item->name, "comment") == 0)
			;
 		else
			errormsg("Invalid table item: %s", item->name);
		}

	fprintf(output, "</table>\n");

	return 0;
	}

int
convert_datasource(xmlNodePtr report, FILE * output, xmlNodePtr datasource)
	{
	xmlNodePtr cur_node;
	char *database, *name, *query = NULL, *old, *data;
	int retval;

	name = xmlGetProp(datasource, "name");
	database = xmlGetProp(datasource, "database");

	for (cur_node = datasource->children; cur_node != NULL;
	        cur_node = cur_node->next)
		{
		if (strcmp(cur_node->name, "text") == 0)
			{
			old = query;
			query = string_extend(query, cur_node->content);
			if (old)
				free(old);
			}
		else if (strcmp(cur_node->name, "data") == 0)
			{
			data = data_retrieve(cur_node);
			old = query;
			query = string_extend(query, data);
			if (old)
				free(old);
			if (data)
				free(data);
			}
		else
			errormsg("Invalid datasource child: %s", cur_node->name);
		}

	retval = datasource_add_raw(name, database, query);

	if (query)
		free(query);

	return retval;
	}

static int
ensure_variable(char *name)
	{
	if (variable_get(name) == NULL)
		{
		errormsg("Required argument %s not defined", name);
		return -1;
		}

	return 0;
	}

/*
   Testing the use of xpath for retrieving node sets: a compromise
   between the use of xslt and good old c code for parsing.
*/
int
check_requirements(xmlDocPtr doc)
	{
	xmlNodePtr node;
	xmlXPathContextPtr xpathCtx; /* xpath context and */
	xmlXPathObjectPtr xpathObj;  /* object pointers */

	xmlNodeSetPtr nodes;			  /* node set */

	xmlChar *xpathExpr;			  /* xpath expression */

	int size;
	int i;

	char *content;

	/*
	   A lot of this is "preliminary" stuff which, if we were using xpath
	   (instead of loops) we'd put in an init function, but for now this just
	   makes this a bit function */

	/* xpath expression to get all 'requires' tags */
	xpathExpr = xmlStrdup("//requires");

	/* Create xpath evaluation context */
	xpathCtx = xmlXPathNewContext(doc);
	if (xpathCtx == NULL)
		{
		errormsg("Error: unable to create new XPath context\n");
		return -1;
		}

	/* Evaluate xpath expression */
	xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);
	if (xpathObj == NULL)
		{
		errormsg("Error: unable to evaluate xpath expression \"%s\"\n",
		         xpathExpr);
		xmlXPathFreeContext(xpathCtx);
		return ( -1);
		}

	/* Nodes set */
	nodes = xpathObj->nodesetval;

	/* The nodes set size */
	size = (nodes) ? nodes->nodeNr : 0;

	/* Go through each node in the node set */
	for (i = 0; i < size; ++i)
		{
		assert(nodes->nodeTab[i]);
		node = nodes->nodeTab[i];			 /* current node in node set */
		if (strcmp(node->name, "requires") == 0)
			{
			char *defaultval;

			defaultval = xmlGetProp(node, "defaultvalue");
			/*
			   type = xmlGetProp(node, "type"); type checking for requires
			   needs to be implemented */
			content = xmlNodeGetContent(node);
			if (!defaultval)
				{
				if (ensure_variable(content) == -1)
					return -1;
				}
			else
				{
				variable_add(content, defaultval);
				}
			}
		}
	return 0;
	}

static int
convert_clearpage(xmlNodePtr report, FILE * output, xmlNodePtr node)
	{
	fprintf(output, "\n\n<clearpage />\n\n");

	return 0;
	}

static int
convert_newpage(xmlNodePtr report, FILE * output, xmlNodePtr node)
	{
	fprintf(output, "\n\n<newpage />\n\n");

	return 0;
	}

int
convert_report(xmlDocPtr doc, FILE * output)
	{
	xmlNodePtr report, node;

	if (doc == NULL)
	     return -1;
	report = xmlDocGetRootElement(doc);

	/* xml_display_node (report); */

	if (check_requirements(doc) != 0)
		return -2;
	fprintf(output, "<?xml version='1.0' encoding='iso-8859-1'?>\n");

	print_tag(output, "report", report, NULL, 0);

	fprintf(output, "\n");

	for (node = report->children; node != NULL; node = node->next)
		{
		if (strcmp(node->name, "datasource") == 0)
			convert_datasource(report, output, node);
		else if (strcmp(node->name, "database") == 0)
			database_add(report, node);
		else if (strcmp(node->name, "table") == 0)
			convert_table(report, output, node);
		else if (strcmp(node->name, "autotable") == 0)
			convert_autotable(report, output, node);
		else if (strcmp(node->name, "label") == 0 ||
		         strcmp(node->name, "spacer") == 0 ||
		         strcmp(node->name, "line") == 0 ||
		         strcmp(node->name, "newpage") == 0 ||
			 strcmp(node->name, "image") == 0)
			convert_generic(report, output, node);
		else if (strcmp(node->name, "requires") == 0)
			;
		else if (strcmp(node->name, "text") == 0)
			;
		else if (strcmp(node->name, "comment") == 0)
			;
		else if (strcmp(node->name, "special") == 0)
			convert_special(output, node);
		else if (strcmp(node->name, "clearpage") == 0)
			convert_clearpage(report, output, node);
		else if (strcmp(node->name, "dataloop") == 0)
			convert_dataloop(report, output, node);
		else if (strcmp(node->name, "title") == 0)
			convert_title(report, output, node);
		else
			{
			errormsg("Invalid report node type: %s", node->name);
			errormsg
			("If you're using entities, make sure you have validation turned on (with the '-d' option)");
			}
		}

	fprintf(output, "</report>\n");

	datasource_clear();

	return 0;
	}
