#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "debug.h"
#include "xml_funcs.h"
#include "datasource.h"

#ifdef WIN32
#include "binding_pg.h"
#endif

static GList *datasources = NULL;
static GList *databases = NULL;

int
database_clear(void)
	{
	GList *cur;
	Database *db;

	for (cur = databases; cur != NULL; cur = cur->next)
		{
		db = cur->data;
#ifndef WIN32		   
		db->binding->pap_disconnect_db(db->db_con);
#else
		pap_disconnect_db(db->db_con);
#endif		   
		g_free(db);
		}

	g_list_free(databases);
	return 0;
	}

int
datasource_remove(DataSource * data)
	{
	if (!data)
		return 0;

	/* debugmsg("Removed datasource: %s", data->name); */
#ifndef WIN32
	if (data->db_query)
		data->db->binding->pap_clearquery(data->db_query);
#else
	if (data->db_query)
		pap_clearquery(data->db_query);
#endif
	g_free(data->name);
	g_free(data->query);

	datasources = g_list_remove(datasources, data);

	g_free(data);

	return 0;
	}

int
datasource_clear(void)
	{
	GList *cur;
	DataSource *data;

	for (cur = datasources; cur != NULL; cur = cur->next)
		{
		data = cur->data;
		g_free(data);
		}

	g_list_free(datasources);

	datasources = NULL;

	return 0;
	}

int
datasource_add_raw(char *name, char *database, char *query)
	{
	DataSource *data;

	/* debugmsg("name: %s, database: %s, query: %s", name, database, query); */

	if (!name || !database || !query)
		{
		if (!name)
			errormsg("Datasource must specify a name");
		if (!database)
			errormsg("Datasource must specify a database");
		if (!query)
			errormsg("Datasource must specify a query");

		return -1;
		}

	if ((data = datasource_get(name)) != NULL)
		{
		datasource_remove(data);
		}

	data = malloc(sizeof(*data));

	data->name = strdup(name);
	data->query = strdup(query);
	data->db = database_get(database);

	if (!data->db)
		{
		errormsg("Invalid database: %s", database);
		free(data);
		return -1;
		}

	data->db_query = NULL;

	datasources = g_list_append(datasources, data);

	return 0;
	}

int
datasource_query(DataSource * data)
	{
	if (!data || !data->db)
		{
		errormsg("Invalid data source");
		return -1;
		}
#ifndef WIN32
	if (data->query && data->db_query &&
	        data->db->binding->pap_query_compare(data->db_query, data->query) == 0)
		{
		data->db->binding->pap_query_reset(data->db_query);
		return 0;
		}
#else
	if (data->query && data->db_query &&
	        pap_query_compare(data->db_query, data->query) == 0)
		{
		pap_query_reset(data->db_query);
		return 0;
		}	   
#endif

#ifndef WIN32
	if (data->db_query)
		data->db->binding->pap_clearquery(data->db_query);

	data->db_query =
	    data->db->binding->pap_doquery(data->db->db_con, data->query);
#else
	if (data->db_query)
		pap_clearquery(data->db_query);

	data->db_query =
	    pap_doquery(data->db->db_con, data->query);
#endif
	if (!data->db_query)
		{
		errormsg("Unable to perform query: %s", data->query);
		free(data);
		return -1;
		}

	return 0;
	}

int
database_add_raw(char *name, char *type, char *provider, char *connectstr)
	{
	Database *db;
#ifndef WIN32	   
	Binding *binding;
#endif
	/* Make sure arguments are valid */
	if (!name || !connectstr || !type)
		{
		if (!name)
			g_error("Database must specify a name");
		if (!connectstr)
			g_error("Database must specifiy a connection");
		if (!type)
			g_error("Database must specify a database type");
		return -1;
		}
	else if ((strcmp(type, "libgda") == 0) && !provider)
		{
		g_error("libgda requires a provider (eg `PostgreSQL'");
		return -1;
		}

#ifndef WIN32
	binding = binding_find(type);
	if (!binding)
		{
		g_error("Unable to find bindings for `%s'", type);
		return -1;
		}

#endif
	db = malloc(sizeof * db);
	db->name = name;
	db->connection = connectstr;
	db->type = type;
	db->provider = provider;
	
#ifndef WIN32
	db->binding = binding;
	db->db_con = db->binding->pap_connect_db(db->connection, db->provider);
#else
	db->binding = NULL;
	db->db_con = pap_connect_db(db->connection, db->provider);
#endif	  
	if (!db->db_con)
		{
		g_free(db);
		g_error("Failed to connect to database");
		return -1;
		}

	databases = g_list_append(databases, db);
	return 0;
	}

int
database_add(xmlNodePtr root, xmlNodePtr database)
	{

	gchar *name, *type, *provider, *connectstr;

	name = xmlGetProp(database, "name");
	type = xmlGetProp(database, "type");
	provider = xmlGetProp(database, "provider");
	connectstr = xmlNodeGetContent(database);
	return database_add_raw(name, type, provider, connectstr);
	}

DataSource *
datasource_get(char *name)
	{
	DataSource *cur;
	GList *w;

	for (w = datasources; w != NULL; w = w->next)
		{
		cur = w->data;
		if (strcmp(cur->name, name) == 0)
			return cur;
		}

	return NULL;
	}

Database *
database_get(char *name)
	{
	Database *cur;
	GList *w;

	for (w = databases; w != NULL; w = w->next)
		{
		cur = w->data;
		if (strcmp(cur->name, name) == 0)
			return cur;
		}

	return NULL;
	}
