#define _GNU_SOURCE
#include <ctype.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "xml_funcs.h"
#include "debug.h"

xmlAttr *
xml_node_has_attr(xmlNode * node, char *name)
	{
	xmlAttr *attr;

	for (attr = node->properties; attr != NULL; attr = attr->next)
		{
		if (strcmp(attr->name, name) == 0)
			return attr;
		}

	return NULL;
	}

char *
xml_node_attr_val(xmlNode * node, char *name)
	{
	xmlAttr *a;

	a = xml_node_has_attr(node, name);

	if (!a)
		return NULL;

	return a->children->content;
	}

int
xml_attr_is_val(xmlAttr * attr, char *value)
	{
	if (strcmp(attr->children->content, value) == 0)
		return 1;

	return 0;
	}

xmlAttr *
xml_node_parent_has_attr(xmlNode * node, char *name)
	{
	xmlAttr *a;
	xmlNode *cur;

	for (cur = node; cur != NULL; cur = cur->parent)
		{
		if ((a = xml_node_has_attr(cur, name)) != NULL)
			return a;
		}

	return NULL;
	}

char *
xml_node_parent_attr(xmlNode * node, char *name)
	{
	xmlAttr *a;

	if ((a = xml_node_parent_has_attr(node, name)) != NULL)
		return a->children->content;

	return NULL;
	}

int
xml_node_parent_attr_val(xmlNode * node, char *name, char *value)
	{
	char *s;

	s = xml_node_parent_attr(node, name);

	if (!s)
		return 0;

	if (strcmp(s, value) == 0)
		return 1;

	return 0;
	}

int
xml_node_has_attr_val(xmlNode * node, char *name, char *value)
	{
	xmlAttr *a;

	a = xml_node_has_attr(node, name);

	if (!a)
		return 0;

	return xml_attr_is_val(a, value);
	}

xmlNode *
xml_node_get_last_child(xmlNode * node)
	{
	xmlNode *cur, *prev = NULL;

	for (cur = node->children; cur != NULL; prev = cur, cur = cur->next)
		;

	return prev;
	}

xmlAttr *
xml_node_get_last_property(xmlNode * node)
	{
	xmlAttr *cur, *prev = NULL;

	for (cur = node->properties; cur != NULL; prev = cur, cur = cur->next)
		;

	return prev;
	}

xmlAttr *
xml_node_add_attr(xmlNode * node, char *attr, char *value)
	{
	xmlAttr *last, *new;
	xmlNode *data;

	if (xml_node_has_attr(node, attr) != NULL)
		{
		errormsg("Node %s already has attribute %s, with value %s",
		         node->name, attr, xml_node_attr_val(node, attr));
		return NULL;
		}

	last = xml_node_get_last_property(node);

	data = malloc(sizeof * data);
	new = malloc(sizeof * new);

	new->name = attr;
	new->children = data;
	new->next = NULL;
	new->parent = node;

	data->name = "text";
	data->content = value;
	data->children = NULL;
	data->properties = NULL;
	data->next = data->prev = NULL;
	data->doc = NULL;
	new->_private = data->_private = NULL;

	new->prev = last;

	if (last)
		last->next = new;
	else
		node->properties = new;

	return new;
	}

int
xml_node_count_children(xmlNode * node, char *childname)
	{
	xmlNode *item;
	int count = 0;

	dassert(node);
	dassert(childname);

	for (item = node->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, childname) == 0)
			count++;
		}

	return count;
	}

static char *
left_trim(char *text)
	{
	char *retval;
	char *pos;
	int start = 0;

	if (!text)
		return NULL;

	retval = malloc(strlen(text) + 1);

	pos = retval;
	for (; *text != '\0'; text++)
		{
		if (!start && isblank(*text))
			continue;
		*pos = *text;
		start = 1;
		pos++;
		}

	*pos = '\0';

	return retval;
	}

static char *
right_trim(char *text)
	{
	char *retval;
	int i;

	if (!text)
		return NULL;

	retval = malloc(strlen(text) + 1);
	strcpy(retval, text);

	for (i = strlen(retval) - 1; i >= 0; i--)
		{
		if (isblank(retval[i]))
			retval[i] = '\0';
		else
			break;
		}
	return retval;
	}

char *
trim(char *text)
	{
	char *val1, *val2;

	val1 = left_trim(text);
	val2 = right_trim(val1);
	if (val1)
		free(val1);

	return val2;
	}

char *
string_extend(char *string1, char *string2)
	{
	char *retval;
	int len = 0;

	if (string1)
		len += strlen(string1);
	if (string2)
		len += strlen(string2);

	retval = malloc(len + 1);
	retval[0] = '\0';

	if (string1)
		{
		strcpy(retval, string1);
		retval[strlen(string1)] = '\0';
		}

	if (string2)
		strcat(retval, string2);

	retval[len] = '\0';

	return retval;
	}

#ifdef WIN32
int
isblank(int c)
	{
	if (c == ' ' || c == '\t')
		return -1;
	return 0;
	}
#endif
