#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <glib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "latex_output.h"
#include "xml_funcs.h"
#include "debug.h"

#define max(a,b) (((a) > (b)) ? (a) : (b))

int
latex_title(FILE * output, xmlNode * title)
	{
	xmlNode *item = NULL;

	/* Is there a title element? */
	if (title == NULL)
		{
		errormsg("There doesn't seem to be a title element, perhaps a "
		         "report name exists?");
		return -1;
		}
	for (item = title->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0)
			latex_label(output, item, 0, 0);
		else
			errormsg("Invalid title item: %s", item->name);
		}
	return 0;
	}

char *
convert_width(char *width)
	{
	char *retval;

	if (!width)
		return NULL;
	if (strchr(width, '%') != 0)
		retval = (char *)g_strdup_printf("%f\\fulltextwidth", 
			(atof(width)) / 100);
	else
		retval = (char *)g_strdup_printf("%s", width);

	return retval;
	}

int
latex_image(FILE * output, xmlNode * image, int border)
	{
	char *imagefile;
	char *width, *height;
	int printed = 0;
	char *tempstr;

	width = xml_node_parent_attr(image, "width");
	height = xml_node_parent_attr(image, "height");

	imagefile = image->children->content;
	fprintf(output, "\\includegraphics");

	if (width)
		{
		tempstr = convert_width(width);
		fprintf(output, "[width=%s", tempstr);
		free(tempstr);
		printed = 1;
		}

	if (height)
		{
		fprintf(output, "%sheight=%s", printed ? ", " : "[", height);
		printed = 1;
		}

	if (printed)
		fprintf(output, "]");

	fprintf(output, "{%s}\n", imagefile);

	return 0;
	}

int
latex_string(FILE * output, char *text)
	{
	int i;
	int placedchar = 0;
	if (!text)
	     return;
	/* perform character translation for latex */
	for (; *text != '\0'; text++)
		{
		/* Trim trailing spaces */
		/* Forward if space character keep occurring */
		for (i = 0; *(text + i); i++)
			{
			if (!isspace(*(text + i)))
				break;
			}
		if (!*(text + i))
			{
			fputc('\n', output);
			return 0;
			}

		if (*text == '\n')
			{
			if (placedchar == 0)
				fprintf(output, "\\hspace{0.001pt}\\\\\n\n");
			else
				fprintf(output, "\\\\\n\n");
			placedchar = 0;
			continue;
			}
		else if (strchr("<",*text))
			fprintf(output, "\\textless{}");
		else if (strchr(">",*text))
			fprintf(output, "\\textgreater{}");
		else if (strchr("~",*text))
			fprintf(output, "\\textasciitilde{}");
		else if (strchr("^",*text))
			fprintf(output, "\\textasciicircum{}");
		else if (strchr("&",*text))
			fprintf(output, "\\&amp;");
		else if (strchr("",*text))
			fprintf(output, "\\`a");	
		else if (*text == '')
			fprintf(output, "\\'a");

/*		else if (strchr("|",*text))
			fprintf(output, "\\verb*|*");  */
		else if (strchr("%{}$#_'", *text))
			fprintf(output, "\\%c", *text);
/*		else if (strchr("()", *text))
			fprintf(output, "$%c$", *text); */
		else if (strchr("\\", *text))
			{
			fprintf(output, "\\textbackslash"); 
/*			if (*(text + 1) == '1')
				fprintf(output, "/"); */
			}

		else
			fprintf(output, "%c", *text);
		placedchar = 1;
		}

	return 0;
	}

int
latex_label(FILE * output, xmlNode * label, int border, int paragraph)
	{
	int isbold, isitalic, isparagraph, boxed;
	char *font;
	char *align, *align_str = NULL;
	char *valign, *valign_str = NULL;
	char *width, *width_str = NULL;
	char *size;
	xmlNode *item;

	dassert(label);

	isbold = xml_node_parent_attr_val(label, "bold", "true");
	isitalic = xml_node_parent_attr_val(label, "italic", "true");
	isparagraph = xml_node_parent_attr_val(label, "paragraph", "true");
	boxed = xml_node_parent_attr_val(label, "boxed", "true");

	font = xml_node_parent_attr(label, "font");
	align = xml_node_parent_attr(label, "align");
	valign = xml_node_parent_attr(label, "valign");
	width = xml_node_parent_attr(label, "width");
	size = xml_node_parent_attr(label, "size");

	if (align)
		{
		if (strcmp(align, "left") == 0)
			align_str = strdup("l");
		else if (strcmp(align, "right") == 0)
			align_str = strdup("r");
		else if (strcmp(align, "center") == 0)
			align_str = strdup("c");
		else if (strcmp(align, "justified") == 0)
			;
		else
			errormsg("Invalid label alignment: %s", align);
		}

	if (isparagraph)
		{
		if (!width)
			{
			errormsg("Paragraph alignment specified without width");
			asprintf(&width_str, "4cm");
			}
		else
			width_str = convert_width(width);

		if (valign && strcmp(valign, "middle") != 0)
			{
			if (strcmp(valign, "top") == 0)
				valign_str = strdup("[t]");
			else if (strcmp(valign, "bottom") == 0)
				valign_str = strdup("[b]");
			else
				errormsg("Invalid label vertical-alignment: " "%s", valign);
			}
		else
			valign_str = strdup("\0");

		fprintf(output, "\\begin{minipage}%s{%s}\n",
		        valign_str, width_str);
		if (align_str)
			{
			switch (*align_str)
				{
			case 'l':
				fprintf(output, "\\begin{flushleft}");
				break;
			case 'r':
				fprintf(output, "\\begin{flushright}");
				break;
			case 'c':
				fprintf(output, "\\begin{center}");
				break;
				}
			}
		}
	else if (align_str)
		{
		switch (*align_str)
			{
		case 'l':
			fprintf(output, "{\\raggedleft ");
			break;
		case 'r':
			fprintf(output, "{\\raggedright ");
			break;
		case 'c':
			fprintf(output, "{\\centering ");
			break;
			}
		}

	if (boxed)
		{
		fprintf(output, "\\fbox{");
		}

	if (font)
		{
		if (size)
			{
			if (!strcmp(size, "big"))
				{
				fprintf(output, "\\fontsize{16}{18pt}\\selectfont");
				}
			else if (!strcmp(size, "large"))
				{
				fprintf(output, "\\fontsize{12}{14pt}\\selectfont");
				}
			else if (!strcmp(size, "small"))
				{
				fprintf(output, "\\fontsize{8}{10pt}\\selectfont");
				}
			else
				{
				errormsg("Invalid font face! %s\n", font);
				}
			}

		if (!strcmp(font, "Palatino"))
			{
			fprintf(output, "\\usefont{T1}{ppl}{%s}{%s} ",
			        isbold ? "b" : "m", isitalic ? "it" : "n");
			}
		else if (!strcmp(font, "Helvetica"))
			{
			fprintf(output, "\\usefont{T1}{phv}{%s}{%s} ",
			        isbold ? "b" : "m", isitalic ? "it" : "n");
			}
		else if (!strcmp(font, "Times"))
			{
			fprintf(output, "\\usefont{T1}{ptm}{%s}{%s} ",
			        isbold ? "b" : "m", isitalic ? "it" : "n");
			}
		else if (!strcmp(font, "Utopia"))
			{
			fprintf(output, "\\usefont{T1}{put}{%s}{%s} ",
			        isbold ? "b" : "m", isitalic ? "it" : "n");
			}
		else if (!strcmp(font, "Lucida"))
			{
			fprintf(output, "\\usefont{T1}{hlh}{%s}{%s} ",
			        isbold ? "b" : "m", isitalic ? "it" : "n");
			}
		else
			{
			errormsg("Invalid font face! %s\n", font);
			exit(EXIT_FAILURE);
			}
		}
	else
		{

		if (isbold)
			fprintf(output, "\\textbf{");
		if (isitalic)
			fprintf(output, "\\textit{");

		if (size)
			{
			if (!strcmp(size, "big"))
				{
				fputs("{\\Large ", output);
				}
			else if (!strcmp(size, "large"))
				{
				fputs("{\\large ", output);
				}
			else if (!strcmp(size, "small"))
				{
				fputs("{\\scriptsize ", output);
				}
			}

		}

	for (item = label->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "text") == 0 && item->type == XML_TEXT_NODE)
			{
			xmlChar *buf =  xmlNodeGetContent(item);
			latex_string(output, buf);
			xmlFree(buf);
			}
		else if (strcmp(item->name, "image") == 0)
			latex_image(output, item, border);
		else if (strcmp(item->name, "comment") == 0)
			;
		else
			errormsg("Invalid label item: %s", item->name);
		}

	if (!font)
		{

		if (size)
			{
			fputs("} %hello\n", output);
			}

		if (isbold)
			fprintf(output, "}");
		if (isitalic)
			fprintf(output, "}");

		}												 /* font */

	if (boxed)
		{
		fputs("}", output);
		}

	if (isparagraph)
		{
		if (align_str)
			switch (*align_str)
				{
			case 'l':
				fprintf(output, "\\end{flushleft}\n");
				break;
			case 'r':
				fprintf(output, "\\end{flushright}\n");
				break;
			case 'c':
				fprintf(output, "\\end{center}\n");
				break;
				}

		fprintf(output, "\\end{minipage}\n");
		}
	else if (align_str)
		{
		fprintf(output, "}");
		}

	if (width_str)
		free(width_str);

	if (align_str)
		free(align_str);

	fprintf(output, "\n");

	return 0;
	}

int
latex_horizontal_spacer(FILE * output, char *width)
	{
	char *width_str;

	if (!width)
		return -1;
	width_str = convert_width(width);
	fprintf(output, "\\hspace*{%s}", width_str);
	free(width_str);
	return 0;
	}

int
latex_cell_multicol(FILE * output, xmlNode * cell, int span, int border)
	{
	char *alignval;
	char align;

	alignval = xmlGetProp(cell, "align");

	if (alignval)
		{
		if (strcmp(alignval, "right") == 0)
			{
			align = 'r';
			}
		else if (strcmp(alignval, "center") == 0)
			{
			align = 'c';
			}
		else
			{
			align = 'l';
			}
		}
	else
		{
		align = 'l';
		}

	fprintf(output, "\\multicolumn{%d}{%s%c%s}{", span, border ? "|" : "", align,
	        border ? "|" : "");
	return 0;
	}

int
latex_cell_multicol_end(FILE * output)
	{
	fprintf(output, "}");
	return 0;
	}

int
latex_line(FILE * output)
	{
	fprintf(output, "\\hline\n");
	return 0;
	}

int
latex_column_sep(FILE * output)
	{
	fprintf(output, "&");
	return 0;
	}

int
latex_row_end(FILE * output)
	{
	fprintf(output, "\\\\\n");
	return 0;
	}

int
latex_table_topline(FILE * output)
	{
	fprintf(output, "\\hline\n");
	return 0;
	}

int
latex_table_bottomline(FILE * output)
	{

	/* fprintf (output, "\\hline\n"); */
	fprintf(output, "\n");
	return 0;
	}

int
latex_table_sep(FILE * output)
	{
	fprintf(output, "\\hline\\hline\n");
	return 0;
	}

int
latex_table_header_start(FILE * output)
	{
	/* fprintf (output, "\\endfirsthead\n"); */
	return 0;
	}

int
latex_table_header_end(FILE * output)
	{
	fprintf(output, "\\endhead\n");
	return 0;
	}

int
latex_table_footer_start(FILE * output, int border)
	{
	return 0;
	}

int
latex_table_footer_end(FILE * output, int border)
	{
	if (border)
		fprintf(output, "\\hline");

	fprintf(output, "\\endfoot\n");
	return 0;
	}

static int
get_number_table_columns(xmlNode * table)
	{
	int total, cur;
	xmlNode *row;

	total = 0;

	for (row = table->children; row != NULL; row = row->next)
		{
		if (strcmp(row->name, "row") == 0)
			{
			cur = xml_node_count_children(row, "cell");
			total = max(total, cur);
			}
		else if (strcmp(row->name, "header") == 0
		         || strcmp(row->name, "footer") == 0)
			{
			cur = get_number_table_columns(row);
			total = max(total, cur);
			}
		}

	return total;
	}

int
latex_table_width(FILE * output, xmlNode * table)
	{
	int total;
	char *width;
	char *width_output;

	width = xmlGetProp(table, "width");
	if (!width)
		return 0;

	width_output = convert_width(width);

	total = get_number_table_columns(table);

	/* errormsg ("width_output: %s", width_output); */
	fprintf(output, "\\multicolumn{%d}{p{%s}}{ }\\kill\n", total, width_output);

	free(width_output);

	return 0;
	}

static int
latex_table_start_alignment(FILE * output, xmlNode * table, int hasboarder)
	{
	int count = 0;
	char *value, align;
	xmlNode *row, *col, *header;

	for (header = table->children; header != NULL; header = header->next)
		{
		if (strcmp(header->name, "header") == 0)
			{
			for (row = header->children; row != NULL; row = row->next)
				{
				if (strcmp(row->name, "row") == 0)
					{
					for (col = row->children; col != NULL;
					        col = col->next)
						{
						if (strcmp(col->name, "cell") == 0)
							{
							if (count == 0)
								fprintf(output, "{");

							value = xml_node_attr_val(col, "align");
							align = 'l';
							if (!value)
								{
								fprintf(output, "%c", align);
								count++;
								continue;
								}
							if (strcmp(value, "right") == 0)
								align = 'r';
							else if (strcmp(value, "left") == 0)
								align = 'l';
							else if (strcmp(value, "center") == 0)
								align = 'c';
							fprintf(output, "%c", align);
							count++;
							}
						}
					}
				}
			}
		}
	if (count)
		fprintf(output, "}");
	return count;
	}

int
latex_table_start(FILE * output, xmlNode * table, int nested, char align)
	{
	int total, cur, colprinted = 0;
	int hasborder;
	char *border;
	int expand;

	border = xmlGetProp(table, "border");
	if (border)
		hasborder = atoi(border);
	else
		hasborder = 0;

	if (xml_node_has_attr(table, "expand"))
		expand = atoi(xmlGetProp(table, "expand"));
	else
		expand = -1;

	total = get_number_table_columns(table);

	if (nested)
		{
		if (align != '\0')
			errormsg("Can't have table alignment in nested tables!\n");
		fprintf(output, "\n\\begin{tabular}");
		}
	else
		{
		fprintf(output, "\n\\begin{longtable}");
		if (align != '\0')
			fprintf(output, "[%c]", align);
		colprinted = latex_table_start_alignment(output, table, hasborder);
		}

#warning "Ignoring the expand option"
	if (colprinted <= 0)
		{
		fprintf(output, "{");
		for (cur = 0; cur < total; cur++)
			fprintf(output, "%s%sl", hasborder ? "|" : "", ""
			        /* cur == (expand - 1) ? "@{\\hspace{\\stretch{1}}}" : "" */
			       );
		if (hasborder)
			fprintf(output, "|");
		fprintf(output, "}\n");
		}

	return 0;
	}

int
latex_table_end(FILE * output, int nested)
	{
	if (nested)
		fprintf(output, "\\end{tabular}\n");
	else
		fprintf(output, "\\end{longtable}\n");
	return 0;
	}

static int
setup_fancy_headers(FILE * output, xmlNode * title0, char *reportname)
	{
	fprintf(output, "\\pagestyle{fancy}\n");
	fprintf(output, "\\fancyhead[L]{}\n");
	fprintf(output, "\\fancyhead[C]{}\n");
	fprintf(output, "\\fancyhead[R]{\\bfseries ");
	if (title0)
		latex_title(output, title0);
	else
		latex_string(output, reportname);
	fprintf(output, "}\n");
	fprintf(output, "\\fancyfoot[L]{}\n");
	fprintf(output, "\\fancyfoot[C]{}\n");
	fprintf(output, "\\fancyfoot[R]{\\thepage\\ of\\ \\pageref{FINALPAGE}}\n");
	return 0;
	}

/*
Francis: This function tries to imitate hash data structure
*/
static char *
get_special_data(xmlNode * special, char *key)
	{
	char *val = NULL;
	char *name;
	xmlNode *curr;

	g_assert(key);

	for (curr = special->children; curr; curr = curr->next)
		{
		if (strcmp(curr->name, "arg") == 0)
			{
			name = xmlGetProp(curr, "name");
			if (name)
				{
				if (strcmp(name, key) == 0)
					{
					val = curr->children->content;
					break;
					}
				}
			else
				{
				errormsg("Invalid arg because of no name attribute value");
				}
			}
		}

	return val;
	}

int
latex_header(FILE * output, xmlNode * report)
	{
	char *reportname;
	char *orientation;
	char *papersize;
	int landscape = 0;
	int letterpaper = 0;

	/* Francis: Sorry to break presentation architecture */
	char *val;
	char defaultfontsize[5] = "10pt";
	xmlNode *curr = NULL;
	xmlNode *special = NULL;
	xmlNode *title0 = NULL;

	reportname = xmlGetProp(report, "name");

	/* 	if (!reportname) */

	/* 		{ */

	/* 		errormsg("Report must have name"); */

	/* 		return -1; */

	/* 		} */

	orientation = xmlGetProp(report, "orientation");
	if (orientation)
		if (strcmp(orientation, "landscape") == 0)
			landscape = 1;

	papersize = xmlGetProp(report, "papersize");
	if (papersize)
		if (strcmp(papersize, "letterpaper") == 0)
			letterpaper = 1;

	for (curr = report->children; curr; curr = curr->next)
		{
		if (strcmp(curr->name, "special") == 0)
			{
			special = curr;
			break;
			}
		if (strcmp(curr->name, "title") == 0)
			{
			title0 = curr;
			break;
			}
		}

	/* 	if (!title0) */
	/* errormsg("Report should have a title, unless it has a name instead."); */
	/* return - 1; */

	if (special)
		{
		if ((val = get_special_data(special, "defaultfontsize")))
			{
			strcpy(defaultfontsize, val);
			}
		}

	fprintf(output, "\\documentclass[%s]{article}\n", defaultfontsize);
	if (landscape)
		{
		if (letterpaper)
			{
			fprintf(output,
			        "\\usepackage[letterpaper, landscape, bottom=2cm, top=3cm, left=2cm, right=2cm]{geometry}\n"
			        "\\def\\printlandscape{\\special{papersize=11in,8.5in}}\n"
			        "\\special{! TeXDict begin /landplus90{true}store end }\n");
			}
		else
			{
			fprintf(output,
			        "\\usepackage[a4paper, landscape, bottom=2cm, top=3cm, left=2cm, right=2cm]{geometry}\n"
			        "\\def\\printlandscape{\\special{papersize=29.7cm,21cm}}\n"
			        "\\special{! TeXDict begin /landplus90{true}store end }\n");
			}
		}
	if (special && (val = get_special_data(special, "basefontpackage")))
		{
		fprintf(output, "\\usepackage{%s}\n", val);
		}

	if (special && (val = get_special_data(special, "margincontrol")))
		{
		gchar **params;

		fprintf(output, "\\usepackage{vmargin}\n");
		fprintf(output, "\\setpapersize[%s]{A4}\n",
		        orientation ? orientation : "portrait");

		params = g_strsplit(val, ",", 8);

		/* fprintf (output,
		   "\\setmarginsrb{1cm}{1cm}{1cm}{1cm}{0pt}{0mm}{0pt}{0mm}\n"); */
		fprintf(output, "\\setmarginsrb{%s}{%s}{%s}{%s}{%s}{%s}{%s}{%s}\n", params[0], 	/* left
		        																														   margin 
		        																														 */
		        params[1], 					 /* top margin */
		        params[2], 					 /* right margin */
		        params[3], 					 /* bottom margin */
		        params[4], 					 /* head height */
		        params[5], 					 /* head separation */
		        params[6], 					 /* foot height */
		        params[7]						 /* foot skip */
		       );
		g_strfreev(params);

		}
	else
		{
		if (!landscape)
			{
			if (letterpaper)
				fprintf(output,
				        "\\usepackage[letterpaper, bottom=2cm, top=3cm, left=2cm, right=2cm]{geometry}\n");
			else
				fprintf(output,
				        "\\usepackage[a4paper, bottom=2cm, top=3cm, left=2cm, right=2cm]{geometry}\n");
			}
		}

	/* if (landscape) { fprintf (output, "\\usepackage{lscape}\n"); fprintf
	   (output, "\\special{landscape}\n"); } */
	fprintf(output, "\\usepackage{longtable}\n");
	fprintf(output, "\\usepackage{graphicx}\n");
	fprintf(output, "\\usepackage{fancyhdr}\n");
	fprintf(output, "\\usepackage[latin1]{inputenc}\n");
	fprintf(output, "\\newlength{\\fulltextwidth}\n");
	fprintf(output, "\\setlength{\\fulltextwidth}{\\textwidth}\n");

	/*
	   fprintf (output, "\\oddsidemargin=0in\n"); */

	setup_fancy_headers(output, title0, reportname);

	fprintf(output, "\\title{");
	if (title0)
		latex_title(output, title0);
	else
		latex_string(output, reportname);
	fprintf(output, "}\n");
	fprintf(output, "\\begin{document}\n");

	/* dvips will automatically print landscape using this... */
	if (landscape)
		fprintf(output, "\\printlandscape\n");

	/* 	fprintf (output, "\\maketitle\n"); */

	return 0;
	}

int
latex_footer(FILE * output)
	{
	fprintf(output, "\\label{FINALPAGE}\n");
	fprintf(output, "\\end{document}");
	return 0;
	}
