#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "debug.h"
#include "xml2latex.h"
#include "latex_output.h"
#include "xml_funcs.h"

static int gen_table(xmlNode * report, FILE * output, xmlNode * table, int nested);

static int
gen_newpage(xmlNode * report, FILE * output, xmlNode * newpage)
	{
	fprintf(output, "\\newpage\n");
	return 0;
	}

static int
gen_line(xmlNode * report, FILE * output, xmlNode * line)
	{
	char *style;
	char *width;
	char *width_str;

	style = xmlGetProp(line, "style");
	style = style ? style : "single";
	width = xmlGetProp(line, "width");
	width = width ? width : "100%";

	fprintf(output, "\\vspace{-1cm}\n");

	width_str = convert_width(width);
	fprintf(output, "\\begin{center}\\begin{tabular}{p{%s}}\n", width_str);

	if (strcmp(style, "single") == 0)
		fprintf(output, "\\hline \\vspace{1pt}");
	else if (strcmp(style, "double") == 0)
		fprintf(output, "\\hline \\hline \\vspace{1pt}");
	else
		errormsg("Invalid line style: %s", style);

	fprintf(output, "\n\\end{tabular}\\end{center}\n");

#warning "Not working out the reverse spacing after a horizontal line correctly"

	fprintf(output, "\\vspace{-1cm}\n");

	free(width_str);

	return 0;
	}

static int
gen_spacer(xmlNode * report, FILE * output, xmlNode * spacer)
	{
	char *style;
	char *size = NULL;
	xmlNode *child;

	style = xmlGetProp(spacer, "style");
	style = style ? style : "vertical";

	for (child = spacer->children; child != NULL; child = child->next)
		{
		if (strcmp(child->name, "text") == 0)
			size = child->content;
		else
			errormsg("Invalid spacer item: %s", child->name);
		}

	if (size == NULL)
		{
		errormsg("Spacer node must have content");
		return -1;
		}

	if (strcmp(style, "vertical") == 0)
		fprintf(output, "\\vspace{%s}\n", size);
	else if (strcmp(style, "horizontal") == 0)
		fprintf(output, "\\hspace{%s}\n", size);
	else
		errormsg("Invalid spacer style: %s", style);

	return 0;
	}


static int
gen_label(xmlNode * report, FILE * output, xmlNode * label, int border, int paragraph)
	{
	latex_label(output, label, border, paragraph);
	return 0;
	}

static int
gen_image(xmlNode * report, FILE * output, xmlNode * image, int border)
	{
	latex_image(output, image, border);
	return 0;
	}

static int
gen_cell(xmlNode * report, FILE * output, xmlNode * cell, int border)
	{
	xmlNode *item;
	char *span;
	char *width;
	int printed = 0;

	span = xmlGetProp(cell, "span");
	if (span)
		{
		latex_cell_multicol(output, cell, atoi(span), border);
		}

	width = xmlGetProp(cell, "width");

	for (item = cell->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0)
			{
			gen_label(report, output, item, border, 0);
			printed = 1;
			}
		else if (strcmp(item->name, "paragraph") == 0)
			{
			gen_label(report, output, item, border, 1);
			printed = 1;
			}
		else if (strcmp(item->name, "image") == 0)
			{
			gen_image(report, output, item, border);
			printed = 1;
			}
		else if (strcmp(item->name, "table") == 0)
			{
			gen_table(report, output, item, 1);
			printed = 1;
			}
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid cell item: %s", item->name);
		}

	if (!printed && width)
		latex_horizontal_spacer(output, width);

	if (span)
		latex_cell_multicol_end(output);

	return 0;
	}

static int
cell_follows(xmlNode * item)
	{
	xmlNode *cell;

	for (cell = item->next; cell != NULL; cell = cell->next)
		if (strcmp(cell->name, "cell") == 0)
			return 1;

	return 0;
	}

static int
gen_row(xmlNode * report, FILE * output, xmlNode * row, int border)
	{
	xmlNode *item;

	for (item = row->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "cell") == 0)
			{
			gen_cell(report, output, item, border);
			if (cell_follows(item))
				latex_column_sep(output);
			}
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid row item: %s", item->name);

		}

	latex_row_end(output);
	if (border)
		fputs("\\hline\n", output);

	return 0;
	}

static int
gen_header(xmlNode * report, FILE * output, xmlNode * header, int border)
	{
	xmlNode *row;

	latex_table_header_start(output);

	for (row = header->children; row != NULL; row = row->next)
		{
		if (strcmp(row->name, "row") == 0)
			gen_row(report, output, row, border);
		else if (strcmp(row->name, "text") == 0)
			;
		else
			errormsg("Invalid header item: %s", row->name);
		}

	if (border)
		latex_table_sep(output);

	latex_table_header_end(output);

	return 0;
	}

static int
gen_footer(xmlNode * report, FILE * output, xmlNode * footer, int border)
	{
	xmlNode *row;
	xmlNode *children = footer ? footer->children : NULL;

	latex_table_footer_start(output, border);

	for (row = children; row != NULL; row = row->next)
		{
		if (strcmp(row->name, "row") == 0)
			gen_row(report, output, row, border);
		else if (strcmp(row->name, "text") == 0)
			;
		else
			errormsg("Invalid footer item: %s", row->name);
		}

	latex_table_footer_end(output, border);

	return 0;
	}
static int
gen_table(xmlNode * report, FILE * output, xmlNode * table, int nested)
	{
	xmlNode *item;
	int border;
	int has_footer = 0;
	char *align;

	align = xmlGetProp(table, "align");

	latex_table_start(output, table, nested, align ? align[0] : '\0');

	if (xml_node_has_attr(table, "border"))
		border = atoi(xmlGetProp(table, "border"));
	else
		border = 0;

	latex_table_width(output, table);

	if (border)
		latex_table_topline(output);

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "header") == 0)
			gen_header(report, output, item, border);
		else if (strcmp(item->name, "footer") == 0)
			{
			has_footer = 1;
			gen_footer(report, output, item, border);
			}
		}

	if (!has_footer && border)
		gen_footer(report, output, NULL, border);

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			gen_row(report, output, item, border);
		else if (strcmp(item->name, "line") == 0)
			latex_line(output);
		else if (strcmp(item->name, "text") == 0 ||
		         strcmp(item->name, "header") == 0 || strcmp(item->name, "footer") == 0)
			;
		else
			errormsg("Invalid table item: %s", item->name);
		}

	if (border)
		latex_table_bottomline(output);

	latex_table_end(output, nested);

	return 0;
	}

static int
gen_clearpage(xmlNode * report, FILE * output, xmlNode * node)
	{
	fprintf(output, "\n\n\\clearpage\n\n");

	return 0;
	}

static int
gen_report(xmlNode * report, FILE * output)
	{
	xmlNode *item;
	latex_header(output, report);

	for (item = report->children; item != NULL; item = item->next)
	  {
	    if (strcmp(item->name, "table") == 0)
	      gen_table(report, output, item, 0);
	    else if (strcmp(item->name, "label") == 0)
	      gen_label(report, output, item, 0, 0);
	    else if (strcmp(item->name, "line") == 0) 
	      gen_line(report, output, item);
	    else if (strcmp(item->name, "spacer") == 0) 
	      gen_spacer(report, output, item);
	    else if (strcmp(item->name, "newpage") == 0) 
	      gen_newpage(report, output, item);
	    else if (strcmp(item->name, "special") == 0)
	      ;
	    else if (strcmp(item->name, "clearpage") == 0) 
	      gen_clearpage(report, output, item);
	    else if (strcmp(item->name, "text") == 0) 
	      ;
	    else if (strcmp(item->name, "title") == 0)
	      ;
	    else 
	      errormsg("Invalid report item: %s", item->name);
	  }

	latex_footer(output);

	return 0;
	}

int
xml_doc_to_latex(xmlDoc * doc, FILE * output)
	{
	xmlNode *report;

	dassert(doc);

	report = xmlDocGetRootElement(doc);

	dassert(report);
	dassert(report->name);

	if (strcmp(report->name, "report") != 0)
		{
		errormsg("Invalid report first item: %s", report->name);
		return -1;
		}

	gen_report(report, output);

	return 0;
	}
