#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "debug.h"
#include "xml2text.h"
#include "xml_funcs.h"

#define max(a,b) (((a)>(b)) ? (a) : (b))

/* Francis: To enable mail label printing */
static int virtual_n_cells = 1;
static int virtual_width_cell = 28;	/* = (80 / 3) - side margins */

static int
output_line(FILE * output, int width)
	{
	int i;

	for (i = 0; i < width; i++)
		fprintf(output, "-");

	fprintf(output, "\n");
	return 0;
	}

static int
label_width(xmlNode * label)
	{
	xmlNode *item;
	int retval = 0;

	for (item = label->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "text") == 0)
			retval += strlen(item->content);
		}

	return retval;
	}

static int
cell_width(xmlNode * cell)
	{
	xmlNode *item;
	int retval = 0;

	for (item = cell->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0)
			retval += label_width(item);
		}

	return retval;
	}

static int
row_ncells(xmlNode * row)
	{
	xmlNode *item;
	int retval = 0;

	for (item = row->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "cell") == 0)
			retval++;
		}

	return retval;
	}

static int
row_width_cell(xmlNode * row, int cell)
	{
	xmlNode *item;
	int cur = 0;

	for (item = row->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "cell") == 0)
			{
			if (cur == cell)
				return cell_width(item);
			else
				cur++;
			}
		}

	return 0;
	}

static int
row_width(xmlNode * row)
	{
	int retval = 0;
	xmlNode *item;
	char *border;
	int hasborder;
	int inc;

	border = xml_node_parent_attr(row, "border");
	hasborder = border ? atoi(border) : 0;
	if (hasborder)
		inc = 3;
	else
		inc = 1;

	for (item = row->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "cell") == 0)
			{
			retval += cell_width(item) + inc;
			}
		}
	return retval;
	}

static int
table_ncells(xmlNode * table)
	{
	xmlNode *item;
	int retval = 0, cur;

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			{
			cur = row_ncells(item);
			retval = max(retval, cur);
			}
		}

	return retval;
	}

static int
table_width_cell(xmlNode * table, int cell)
	{
	xmlNode *item;
	int retval = 0, cur;

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "header") == 0 || strcmp(item->name, "footer") == 0)
			{
			cur = table_width_cell(item, cell);
			retval = max(retval, cur);
			}
		else if (strcmp(item->name, "row") == 0)
			{
			cur = row_width_cell(item, cell);
			retval = max(retval, cur);
			}
		}

	return retval;
	}

static int
table_width(xmlNode * table)
	{
	xmlNode *item;
	int retval = 0, cur;

	for (item = table->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "header") == 0 || strcmp(item->name, "footer") == 0)
			{
			cur = table_width(item);
			retval = max(retval, cur);
			}
		else if (strcmp(item->name, "row") == 0)
			{
			cur = row_width(item);
			retval = max(retval, cur);
			}
		}

	return retval;
	}

static int
report_width(xmlNode * report)
	{
	xmlNode *item;
	int retval = 0, cur;

#warning "doesn't seem to get the right width"

	for (item = report->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "table") == 0)
			{
			cur = table_width(item);
			retval = max(retval, cur);
			}
		}

	return retval;
	}

static int
convert_label(xmlNode * report, FILE * output, xmlNode * label, int *curwidth)
	{
	xmlNode *item;

	for (item = label->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "text") == 0)
			{
			*curwidth = strlen(item->content);
			fprintf(output, "%s", item->content);
			}
		else if (strcmp(item->name, "image") == 0)
			;
		else
			errormsg("Invalid label item: %s", item->name);
		}

	return 0;
	}

static int
convert_cell(xmlNode * report, FILE * output, xmlNode * cell, int cellwidth)
	{
	int curwidth=0;
	xmlNode *item;
	
	for (item = cell->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "label") == 0)
			convert_label(report, output, item, &curwidth);
		else if (strcmp(item->name, "text") == 0 || strcmp(item->name, "image") == 0)
			;
		else
			errormsg("Invalid cell item: %s", item->name);
		if (cellwidth)
			{
			cellwidth = cellwidth - curwidth;
			if (item->next == NULL)
				fprintf(output, "%-*s", cellwidth, "");
			} 
		}

	return 0;
	}

/**
 * convert_virtual_cells:
 *
 * Francis: This is a hacky extension to enable virtual cell Procedure:
 * (1) Get ahead 'n' number of rows. 
 * (2) Since for each row will
 * contain only one cell
 *
 **/
static int
convert_virtual_cells(xmlNode * report, FILE * output, xmlNode * cur_row, int ncells)
	{
	int i;
	xmlNode *rows[10];
	xmlNode *cells[10];
	xmlNode *labels[10];

	xmlNode *row_item;

	/* Memorise the 'n' number of rows ahead. */
	row_item = cur_row;
	for (i = 0; i < ncells; i++)
		{

		/* Skip space text node */
		for (; row_item && strcmp(row_item->name, "row"); row_item = row_item->next)
			;

		/* If we reach the end of processing... */
		if (!row_item)
			{
			ncells = i;
			break;
			}
		rows[i] = row_item;

		row_item = row_item->next;			 /* If you miss this line ... */
		}

	/* 	for (i = 0; i < ncells; i++) { */
	/* 		printf ("%p\n", rows[i]); */
	/* 	} */
	/* 	exit (1); */

	/* Set to the first <cell> of each row. This is to avoid spaces */
	for (i = 0; i < ncells; i++)
		{
		xmlNode *cell_item;

		/* Skip space text node */
		for (cell_item = rows[i]->children;
		        cell_item && strcmp(cell_item->name, "cell"); cell_item = cell_item->next)
			;

		g_assert(cell_item);
		cells[i] = cell_item;
		}

	/* 	for (i = 0; i < ncells; i++) { */

	/* 		printf ("%p\n", cells[i]); */

	/* 	} */

	/* 	exit (1); */

	/* Set to the first <label> of each cell. This is to avoid spaces */
	for (i = 0; i < ncells; i++)
		{
		xmlNode *label_item;

		/* Skip space text node */
		for (label_item = cells[i]->children;
		        label_item && strcmp(label_item->name, "label"); label_item = label_item->next)
			;

		g_assert(label_item);
		labels[i] = label_item;
		}

	/* 	for (i = 0; i < ncells; i++) { */

	/* 		printf ("%p\n", labels[i]); */

	/* 	} */

	/* 	exit (1); */

	while (labels[0])
		{
		/* For 'n' number of columns (= rows or cells) */
		for (i = 0; i < ncells; i++)
			{

			xmlNode *item;

			for (item = labels[i]->children; item; item = item->next)
				{

				char content_restricted[50];
				int j;

				strncpy(content_restricted, item->content, virtual_width_cell);

				content_restricted[virtual_width_cell - 1] = '\0';

				for (j = 0; j < virtual_width_cell; j++)
					{
					if (content_restricted[j] == '\n')
						{
						content_restricted[j] = ' ';
						}
					}

				if (strcmp(item->name, "text") == 0)
					{
					if (virtual_width_cell)
						fprintf(output, "%-*s", virtual_width_cell, content_restricted);
					else
						fprintf(output, "%s", content_restricted);
					}
				else if (strcmp(item->name, "image") == 0)
					;
				else
					errormsg("Invalid label item: %s", item->name);
				}
			}

		for (i = 0; i < ncells; i++)
			{
			xmlNode *label_item;

			/* Skip space text node */
			for (label_item = labels[i]->next;
			        label_item && strcmp(label_item->name, "label"); label_item = label_item->next)
				;

			labels[i] = label_item;
			}
		fprintf(output, "\n");
		}
	errormsg("Code has been disabled");
	fprintf(output, "\n\n\n\n\n\n");

	return 0;
	}

static int
convert_row(xmlNode * report, FILE * output, xmlNode * row, int ncells, int *widths)
	{
	xmlNode *item;
	int printed;
	int hasborder;
	char *border;
	int cur = 0, curwidth;

	border = xml_node_parent_attr(row, "border");
	hasborder = border ? atoi(border) : 0;

	if (hasborder)
		fprintf(output, "| ");

	for (item = row->children; item != NULL; item = item->next)
		{

		printed = 0;
		if (strcmp(item->name, "cell") == 0)
			{
			if (cur > ncells)
				curwidth = 0;
			else
				curwidth = widths[cur];
			convert_cell(report, output, item, curwidth);
			printed = 1;
			cur++;
			}
		else if (strcmp(item->name, "spacer") == 0)
			{
			fprintf(output, "\n");
			}
		else if (strcmp(item->name, "text") == 0)
			;
		else if (strcmp(item->name, "newpage") == 0)
			;
/*			{
			fprintf(output,"\n");
			} */
		else
			errormsg("Invalid row item: %s", item->name);

		if (printed)
			fprintf(output, "%s", hasborder ? " | " : " ");
		}

	fprintf(output, "\n");

	return 0;
	}

static int
convert_header(xmlNode * report, FILE * output, xmlNode * header, int ncells, int *widths)
	{
	xmlNode *item;

	for (item = header->children; item != NULL; item = item->next)
		{
		if (strcmp(item->name, "row") == 0)
			convert_row(report, output, item, ncells, widths);
		else if (strcmp(item->name, "text") == 0)
			;
		else
			errormsg("Invalid header item: %s", item->name);
		}

	return 0;
	}

static int
convert_table(xmlNode * report, FILE * output, xmlNode * table)
	{
	xmlNode *item;
	int hasborder;
	int printed;
	char *border;
	int ncells, i;
	int *widths;

	border = xml_node_attr_val(table, "border");
	hasborder = border ? atoi(border) : 0;

	if (hasborder)
		output_line(output, table_width(table));

	ncells = table_ncells(table);

	/* Francis: The virtual ncell is my hack to extend mail labeler printing. To simplify situation assumptions have
	   been made. (1) row has only 1 cell consist of several <data>s (2) ncells is forced to have value of virtual ncell 
	 */
	if (ncells == 1 && virtual_n_cells > 1)
		{
		ncells = virtual_n_cells;
		}
	else if (ncells > 1 && virtual_n_cells > 1)
		{
		fprintf(stderr, "Virtual cell cannot be activated because" "row contains more than 1 cell\n");
		exit(1);
		}

	widths = malloc(ncells * sizeof(int));

	/*	for (i = 0; i < ncells; i++)
			widths[i] = table_width_cell (table, i);*/
	for (i = 0; i < ncells; i++)
		{
		if (virtual_n_cells > 1)
			{
			/* Francis: For simplicity, every cell has same width */
			widths[i] = virtual_width_cell;
			}
		else
			{
			widths[i] = table_width_cell(table, i);
			}
		}

	printed = 0;
	for (item = table->children; item != NULL; item = item->next)
		if (strcmp(item->name, "header") == 0)
			{
			convert_header(report, output, item, ncells, widths);
			printed = 1;
			}

	if (hasborder && printed)
		output_line(output, table_width(table));

	for (item = table->children; item != NULL; item = item->next)
		{

		/* Francis: To process virtual cell feature, bypass the original and normal code below. */
		if (virtual_n_cells > 1)
			{
			int i;

			/* This will consume 'ncells' number of row */
			convert_virtual_cells(report, output, item, ncells);
			/* Forward 'ncells' number of rows */
			for (i = 0; i < ncells && item; i++)
				for (item = item->next; item && strcmp(item->name, "row"); item = item->next)
					;

			if (!item)
				return 0;
			else
				continue;
			}

		if (strcmp(item->name, "row") == 0)
			convert_row(report, output, item, ncells, widths);
		else if (strcmp(item->name, "text") == 0 ||
		         strcmp(item->name, "header") == 0 || strcmp(item->name, "footer") == 0)
			;
		else
			errormsg("Invalid table item: %s", item->name);
		}

	for (item = table->children; item != NULL; item = item->next)
		if (strcmp(item->name, "footer") == 0)
			convert_header(report, output, item, ncells, widths);

	if (hasborder)
		output_line(output, table_width(table));

	return 0;
	}

static int
convert_report(FILE * output, xmlNode * report)
	{
	xmlNode *child;

	for (child = report->children; child != NULL; child = child->next)
		{
		if (strcmp(child->name, "table") == 0)
			convert_table(report, output, child);
		else if (strcmp(child->name, "line") == 0)
			output_line(output, report_width(report));
		else if (strcmp(child->name, "newpage") == 0)
			fprintf(output, "%c", 12);
		else if (strcmp(child->name, "spacer") == 0)
#warning not handling spacers correctly

			;
		else if (strcmp(child->name, "title") == 0)
			;
		else if (strcmp(child->name, "text") == 0)
			;
		else
			errormsg("Invalid report item: %s", child->name);
		}

	return 0;
	}

int
xml_doc_to_text(xmlDoc * doc, FILE * output)
	{
	xmlNode *report;

	g_assert(doc);

	report = xmlDocGetRootElement(doc);

	dassert(report);
	dassert(report->name);

/*	report = doc->children;
        report = xmlDocGetRootElement(doc);	   
	dassert(report);
	dassert(report->name); */
	   
        if (strcmp(report->name, "report") != 0)
	     {
	     errormsg("Invalid report first item: %s", report->name);
	     return -1;
	     }
	return convert_report(output, report);
	}
